/**
  *  \file planetscentral/format/format.cc
  *  \brief Format - Main Entry
  */

#include <stdexcept>
#include <memory>
#include <fcntl.h>
#include <signal.h>
#include <stdio.h>
#include "cpluslib/string.h"
#include "int/args.h"
#include "int/array.h"
#include "int/data.h"
#include "planetscentral/dbclient/connection.h"
#include "planetscentral/format/packer.h"
#include "planetscentral/format/simpacker.h"
#include "planetscentral/format/specpacker.h"
#include "planetscentral/format/stringpacker.h"
#include "planetscentral/osutil.h"
#include "server/config.h"
#include "server/jsonin.h"
#include "server/net.h"
#include "util/misc.h"
#include "server/jsonout.h"
#include "int/hash.h"

static string_t arg_listen_host = "127.0.0.1";
static string_t arg_listen_port = "6665";
static int arg_threads = 5;
Mutex output_mutex;

static void
processConfig(string_t key, string_t value)
{
    if (key == "FORMAT.HOST") {
        /* @q Format.Host:Str (Config)
           Listen address for the Format instance. */
        arg_listen_host = value;
    } else if (key == "FORMAT.PORT") {
        /* @q Format.Port:Int (Config)
           Port number for the Format instance. */
        arg_listen_port = value;
    } else if (key == "FORMAT.THREADS") {
        /* @q Format.Threads:Int (Config)
           Number of threads (=maximum number of parallel connections) */
        stringValue(value, arg_threads);
    }
}

/**************************** Format Dispatch ****************************/

static Packer*
createPacker(const string_t name)
{
    /* @type Format
       Specifies the data format to work on.
       Could be one of:
       - %string: data is a single string
       - %engspec: "engspec.dat" file (array of engine specifications)
       - %torpspec: "torpspec.dat" file (array of torpedo specifications)
       - %beamspec: "beamspec.dat" file (array of beam weapon specifications)
       - %hullspec: "hullspec.dat" file (array of hull specifications)
       - %truehull: "truehull.dat" file (player/hull assignments)
       - %sim: "*.ccb" file (battle simulation) */
    if (name == "string") {
        return new StringPacker();
    } else if (name == "engspec") {
        return new EnginePacker();
    } else if (name == "torpspec") {
        return new TorpedoPacker();
    } else if (name == "beamspec") {
        return new BeamPacker();
    } else if (name == "hullspec") {
        return new HullPacker();
    } else if (name == "truehull") {
        return new TruehullPacker();
    } else if (name == "sim") {
        return new SimPacker();
    } else {
        return 0;
    }
}

/***************************** JsonFormatter *****************************/

struct JsonFormatter : public JsonWriter {
 public:
    JsonFormatter();

    virtual void writeLine(const string_t& line);

    const string_t& getString();

    void write(IntValue* v);

 private:
    string_t accum;
};

inline
JsonFormatter::JsonFormatter()
{ }

void
JsonFormatter::writeLine(const string_t& line)
{
    accum += line;
    accum += '\n';
}

inline const string_t&
JsonFormatter::getString()
{
    return accum;
}

void
JsonFormatter::write(IntValue* v)
{
    if (v == 0) {
        writeNull();
    } else if (IntIntValue* iv = dynamic_cast<IntIntValue*>(v)) {
        // FIXME: do we want to distinguish int/bool?
        writeInt(iv->getValue());
    } else if (IntStringValue* sv = dynamic_cast<IntStringValue*>(v)) {
        writeString(sv->getValue());
    } else if (IntArray* av = dynamic_cast<IntArray*>(v)) {
        IntDataSegment& seg = av->getData()->content;
        startArray();
        for (std::size_t i = 0, n = seg.size(); i < n; ++i) {
            write(seg[i]);
        }
        endArray();
    } else if (IntHash* hv = dynamic_cast<IntHash*>(v)) {
        IntHashData& seg = *hv->getData();
        startObject();
        for (std::size_t i = 0, n = seg.getSize(); i < n; ++i) {
            writeHashElem(seg.getName(i));
            write(seg.get(i));
        }
        endObject();
    } else {
        // unknown
        throw std::runtime_error("Invalid value");
    }
}

/****************************** FormatConfig *****************************/

struct FormatConfig {
    std::auto_ptr<Packer> packer;
    bool json;
    IntValue* data;

    FormatConfig()
        : packer(),
          json(false),
          data(0)
        { }
    void configure(IntArgBlock& args);
};

void
FormatConfig::configure(IntArgBlock& args)
{
    // Must have at least two args
    args.checkArgumentCountAtLeast(2);

    // Check format
    packer.reset(createPacker(toStringResult(args.getNext())));
    if (!packer.get()) {
        throw std::runtime_error("Invalid format");
    }

    // Remember data
    data = args.getNext();

    // Optional args
    while (args.getNumArgs() > 0) {
        string_t kw(strUCase(toStringResult(args.getNext())));
        if (kw == "FORMAT") {
            args.checkArgumentCountAtLeast(1);
            string_t fmt(toStringResult(args.getNext()));
            if (fmt == "json") {
                json = true;
            } else if (fmt == "obj") {
                json = false;
            } else {
                throw std::runtime_error("Invalid data type");
            }
        } else if (kw == "CHARSET") {
            args.checkArgumentCountAtLeast(1);
            packer->setCharset(toStringResult(args.getNext()));
        } else {
            throw std::runtime_error("Invalid option");
        }
    }
}

/**************************** FormatConnection ***************************/

/** Talk Connection. Handles per-connection state. */
class FormatConnection : public Connection {
 public:
    FormatConnection();

    void handleRequest();
    void handleCommand(const string_t& cmd, IntArgBlock& args);
};

/** Constructor. */
FormatConnection::FormatConnection()
    : Connection()
{ }

/** Handle a single request. */
void
FormatConnection::handleRequest()
{
    // Read request
    std::auto_ptr<IntValue> req(readObject(true));
    IntArray* a = dynamic_cast<IntArray*>(req.get());
    if (!a || a->getData()->getDimension(0) == 0) {
        sendError("400 Syntax error");
        return;
    }
    const IntDataSegment& seg = a->getData()->content;

    // Execute request. No logging (will be too complex to log most of the time anyway).
    string_t cmd = strUCase(IntValue::toStringOf(seg[0], false));
    IntArgBlock args(seg, 1, seg.size()-1);
    handleCommand(cmd, args);
}

/** Handle a command.
    \param cmd Command
    \param args Arguments */
void
FormatConnection::handleCommand(const string_t& cmd, IntArgBlock& args)
{
    // Command dispatcher
    if (cmd == "PING") {
        /* @q PING (Format Command)
           Alive test.
           @retval Str "PONG". */
        sendSuccess("PONG");
    } else if (cmd == "HELP") {
        /* @q HELP (Format Command)
           @retval Str Help page. */
        IntStringValue sv("Commands:\n"
                          "  PING\n"
                          "  HELP\n"
                          "  PACK type data [options]\n"
                          "  UNPACK type data [options]\n"
                          "\n"
                          "Options:\n"
                          "  CHARSET cs\n"
                          "  FORMAT {json|obj}\n");
        sendObject(&sv);
    } else if (cmd == "PACK") {
        /* @q PACK out:Format, data, [FORMAT in:Str, CHARSET cs:Str] (Format Command)
           Pack %data into a blob.
           - %in="obj" (default): %data is a RESP object.
           - %in="json": %data is a {@type Str|string} containing a JSON object.

           The character set defines the character encoding in the returned blob and defaults to Latin-1.

           @rettype Blob
           @argtype Str
           @see UNPACK (Format Command) */
        FormatConfig cfg;
        cfg.configure(args);
        if (cfg.json) {
            // Make a tokenizer
            JsonTokenizer tok(toStringResult(cfg.data));
            tok.readNext();

            // Parse
            std::auto_ptr<IntValue> tree(readTree(tok));
            if (tok.getCurrent() != JsonTokenizer::kEOF) {
                throw std::runtime_error("Parse error");
            }

            // Do it
            IntStringValue sv(cfg.packer->pack(tree.get()));
            sendObject(&sv);
        } else {
            // Tree is given as-is
            IntStringValue sv(cfg.packer->pack(cfg.data));
            sendObject(&sv);
        }
    } else if (cmd == "UNPACK") {
        /* @q UNPACK in:Format, data:Blob, [FORMAT out:Str, CHARSET cs:Str] (Format Command)
           Unpack data from a blob.
           - %out="obj" (default): produce a RESP object.
           - %out="json": produce a {@type Str|string} containing a JSON object.

           The character set defines the character encoding in %data and defaults to Latin-1.

           @rettype Str
           @see PACK (Format Command) */
        FormatConfig cfg;
        cfg.configure(args);
        
        // Convert
        // FIXME: this copies the string. Can we avoid that?
        std::auto_ptr<IntValue> tree(cfg.packer->unpack(toStringResult(cfg.data)));
        if (cfg.json) {
            JsonFormatter fmt;
            fmt.write(tree.get());
            fmt.flush();

            IntStringValue sv(fmt.getString());
            sendObject(&sv);
        } else {
            sendObject(tree.get());
        }
    } else {
        sendError("400 Unknown command");
    }
}

static int server_socket;

static void*
serverLoop(void*)
{
    while (1) {
        FormatConnection conn;
        if (!conn.accept(server_socket)) {
            perror("accept");
            sleep(10);
        } else {
            try {
                while (!conn.isAtEnd()) {
                    try {
                        conn.handleRequest();
                    }
                    catch (std::exception& e) {
                        {
                            MutexGuard g(output_mutex);
                            printf("[cmd] [exception] %s\n", e.what());
                        }
                        conn.sendError(e.what());
                    }
                    conn.flush();
                }
            }
            catch (std::exception& e) {
                printf("c2format: %s\n", e.what());
            }
            conn.close();
        }
    }
}

int main()
{
    /* Unbuffer stdout */
    setvbuf(stdout, 0, _IOLBF, 0);

    /* Configuration */
    loadConfiguration(processConfig);

    /* Listen */
    server_socket = Bind().withAddress(arg_listen_host, arg_listen_port);
    if (server_socket < 0) {
        fprintf(stderr, "[main] unable to bind socket\n");
        return 1;
    }

    /* OK, main loop */
    printf("[main] listening...\n");
    signal(SIGPIPE, SIG_IGN);
    fcntl(server_socket, F_SETFD, FD_CLOEXEC);
    listen(server_socket, 10);

    /* Create some threads. Create one less than configured; the main
       thread will also serve requests. */
    for (int32_t i = 1; i < arg_threads; ++i) {
        pthread_t handle;
        if (pthread_create(&handle, 0, serverLoop, 0) < 0) {
            perror("[main] unable to create thread");
            return 1;
        }
    }
    serverLoop(0);
    close(server_socket);
}
