/**
  *  \file planetscentral/net/encoding.cc
  *  \brief Networking: Encoding and Decoding
  */

#include <cstring>
#include "planetscentral/net/encoding.h"

static const char base64_alphabet[] = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
static const char hex_alphabet[] = "0123456789ABCDEF";

static int
hexValue(char c)
{
    if (c >= '0' && c <= '9')
        return c - '0';
    if (c >= 'A' && c <= 'F')
        return c - 'A' + 10;
    if (c >= 'a' && c <= 'f')
        return c - 'a' + 10;
    return -1;
}

static int
b64Value(char c)
{
    const char* p = std::strchr(base64_alphabet, c);
    if (p != 0)
        return p - base64_alphabet;
    else
        return 0;
}

/** Encode Base64. */
string_t
encodeBase64(const string_t input)
{
    string_t result;
    std::size_t index = 0;
    result.reserve((input.size() + 2) / 3 * 4);
    while (index + 3 <= input.size()) {
        unsigned long value = 0;
        value += 65536*(unsigned char) input[index++];
        value += 256*(unsigned char) input[index++];
        value += (unsigned char) input[index++];
        result += base64_alphabet[(value >> 18) & 63];
        result += base64_alphabet[(value >> 12) & 63];
        result += base64_alphabet[(value >> 6) & 63];
        result += base64_alphabet[(value) & 63];
    }

    unsigned long value = 0;
    switch (input.size() - index) {
     case 2:
        value += 65536*(unsigned char) input[index++];
        value += 256*(unsigned char) input[index++];
        result += base64_alphabet[(value >> 18) & 63];
        result += base64_alphabet[(value >> 12) & 63];
        result += base64_alphabet[(value >> 6) & 63];
        result += '=';
        break;
     case 1:
        value += 65536*(unsigned char) input[index++];
        result += base64_alphabet[(value >> 18) & 63];
        result += base64_alphabet[(value >> 12) & 63];
        result += '=';
        result += '=';
        break;
     default:
        break;
    }
    return result;
}

/** Decode Base64. */
string_t
decodeBase64(const string_t input)
{
    string_t result;

    // Trim trailing equal signs. Normal base64 stuff has
    // lengths divisible by four, implementing it this way
    // has the added benefit of getting robustness for
    // truncated streams for free.
    string_t::size_type limit = input.size();
    while (limit > 0 && input[limit-1] == '=')
        --limit;

    result.reserve(limit * 3 / 4 + 1);

    string_t::size_type index = 0;
    while (index < limit) {
        unsigned long value = 0;
        int nch = 0;
        switch (limit - index) {
         default:
            value += b64Value(input[index+3]);
            ++nch;
         case 3:
            value += b64Value(input[index+2]) << 6;
            ++nch;
         case 2:
            value += b64Value(input[index+1]) << 12;
            ++nch;
         case 1:
            value += b64Value(input[index+0]) << 18;
        }
        if (nch > 0)
            result.append(1, char((value >> 16) & 255));
        if (nch > 1)
            result.append(1, char((value >> 8) & 255));
        if (nch > 2)
            result.append(1, char((value) & 255));

        index += 4;
    }
    return result;
}

/** Encode URL-Encoding (percent-escaping). */
string_t
encodeUrl(string_t input)
{
    for (string_t::size_type i = 0; i < input.size(); ++i) {
        unsigned char u = input[i];
        if (u >= 0x80 || u < 0x20 || u == '%' || u == '&' || u == '+') {
            char tmp[2];
            input[i] = '%';
            tmp[0] = hex_alphabet[(u >> 4) & 15];
            tmp[1] = hex_alphabet[u & 15];
            input.insert(i+1, tmp, 2);
            i += 2;
        } else if (u == ' ') {
            input[i] = '+';
        }
    }
    return input;
}

/** Decode URL-Encoding (percent-escaping). */
string_t
decodeUrl(string_t input)
{
    string_t::size_type n = 0;
    while ((n = input.find('+', n)) != input.npos) {
        input[n] = ' ';
        ++n;
    }
    n = 0;
    while ((n = input.find('%', n)) != input.npos && input.size() - n >= 3) {
        int a = hexValue(input[n+1]);
        int b = hexValue(input[n+2]);
        if (a >= 0 && b >= 0) {
            input.erase(n+1, 2);
            input[n] = char(16*a + b);
        }
        ++n;
    }
    return input;
}

/** Encode quoted-printable. */
string_t
encodeQP(string_t input)
{
    for (string_t::size_type i = 0; i < input.size(); ++i) {
        unsigned char u = input[i];
        if (u >= 0x80 || u < 0x20 || u == '=') {
            char tmp[2];
            input[i] = '=';
            tmp[0] = hex_alphabet[(u >> 4) & 15];
            tmp[1] = hex_alphabet[u & 15];
            input.insert(i+1, tmp, 2);
            i += 2;
        }
    }
    return input;
}

/** Decode quoted-printable. */
string_t
decodeQP(string_t input)
{
    string_t::size_type n = 0;
    while ((n = input.find('=', n)) != input.npos && input.size() - n >= 3) {
        int a = hexValue(input[n+1]);
        int b = hexValue(input[n+2]);
        if (a >= 0 && b >= 0) {
            input.erase(n+1, 2);
            input[n] = char(16*a + b);
        }
        ++n;
    }
    return input;
}


static bool
mustEncode(const string_t& word)
{
    for (string_t::size_type i = 0; i < word.size(); ++i) {
        unsigned char u = word[i];
        if (u >= 0x80 || u < 0x20 || u == '?') {
            return true;
        }
    }
    return false;
}

/** Encode MIME header.
    \param input text to encode
    \param charset character set */
string_t
encodeMimeHeader(string_t input, string_t charset)
{
    // FIXME: RFC 2047 places some pretty tight limits on the format of lines containing
    // encoded words:
    // - max 75 chars per encoded word
    // - max 76 chars per line containing an encoded word
    // Since we don't see complete lines, we only try to enforce the per-word limit.
    // "65" is 75 minus "=" and "?", minus roundoff errors.
    string_t::size_type maxCharsPerWord = (65 - charset.size()) * 3/4;
    string_t::size_type n = 0;
    string_t::size_type p;
    string_t result;
    while ((p = input.find_first_not_of(" \t\r\n", n)) != string_t::npos) {
        // Copy run of space characters
        result.append(input, n, p-n);

        // Find next word
        n = input.find_first_of(" \t\r\n", p);
        if (n == string_t::npos) {
            n = input.size();
        }
        string_t word(input, p, n-p);
        if (mustEncode(word)) {
            string_t::size_type i = 0;
            while (i < word.size()) {
                string_t::size_type n = word.size() - i;
                if (n > maxCharsPerWord) {
                    n = maxCharsPerWord;
                }
                result += "=?";
                result += charset;
                result += "?B?";
                result += encodeBase64(string_t(word, i, n));
                result += "?=";
                i += n;
                if (i < word.size()) {
                    result += "\r\n ";
                }
            }
        } else {
            result += word;
        }
    }
    return result;
}
