/**
  *  \file planetscentral/nntp/iothread.cc
  */

#include <stdexcept>
#include <sys/fcntl.h>
#include <sys/select.h>
#include <unistd.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include "planetscentral/nntp/iothread.h"
#include "planetscentral/nntp/nntp.h"
#include "planetscentral/nntp/worker.h"

namespace {
    enum { Read, Write };
};

IOThread::IOThread()
    : handle(),
      started(false),
      terminating(false),
      workers(),
      new_workers(),
      mutex()
{
    // Create wake pipe
    if (pipe(wake_pipe) != 0) {
        throw std::runtime_error("Unable to create pipe");
    }
    if (wake_pipe[Read] >= FD_SETSIZE) {
        throw std::runtime_error("Got an unusable fd for pipe");
    }
    fcntl(wake_pipe[Read], F_SETFL, O_NONBLOCK);
    fcntl(wake_pipe[Write], F_SETFL, O_NONBLOCK);
}

IOThread::~IOThread()
{
    stop();
    close(wake_pipe[Read]);
    close(wake_pipe[Write]);
}

void
IOThread::start()
{
    started = true;
    pthread_create(&handle, 0, entry, this);
}

void
IOThread::stop()
{
    if (started) {
        terminating = true;
        (void) write(wake_pipe[Write], "", 1);
        pthread_join(handle, 0);
        started = false;
        terminating = false;
    }
}

void
IOThread::addWorker(int fd, Worker* worker)
{
    // Make it nonblocking
    fcntl(fd, F_SETFL, O_NONBLOCK);

    // Post it
    {
        MutexGuard g(mutex);
        Info& p = *new_workers.push_back(new Info());
        p.worker.reset(worker);
        p.fd = fd;
        p.status = 0;
    }

    // Wake worker
    (void) write(wake_pipe[Write], "", 1);
}

void*
IOThread::entry(void* p)
{
    static_cast<IOThread*>(p)->run();
    return 0;
}

void
IOThread::run()
{
    char buffer[1024];
    const char* ptr;
    while (1) {
        // Discard old workers
        processFinishedWorkers();

        // Acknowledge new workers
        processNewWorkers();

        // Select loop
        fd_set rset, wset;
        int max;
        FD_ZERO(&rset);
        FD_ZERO(&wset);
        max = 0;

        // Add wake fd
        FD_SET(wake_pipe[Read], &rset);
        if (wake_pipe[Read] >= max) {
            max = wake_pipe[Read]+1;
        }

        // Add workers
        for (std::size_t i = 0; i < workers.size(); ++i) {
            Info& p = *workers[i];
            bool use = false;
            if ((p.status & EndOfInput) == 0) {
                FD_SET(p.fd, &rset);
                use = true;
            }
            if (p.worker->getBytesToSend(ptr) != 0) {
                FD_SET(p.fd, &wset);
                use = true;
            }
            if (use && p.fd >= max) {
                max = p.fd+1;
            }
        }

        // Do it
        int result = select(max, &rset, &wset, 0, 0);
        if (result < 0) {
            MutexGuard g(output_mutex);
            printf("[io] [error] select error %d (%s)\n", errno, strerror(errno));
            if (errno != EINTR) {
                break;
            }
        } else if (terminating) {
            break;
        } else {
            if (FD_ISSET(wake_pipe[Read], &rset)) {
                // Read pipe to clear status
                (void) read(wake_pipe[Read], buffer, sizeof(buffer));
            }
            for (std::size_t i = 0; i < workers.size(); ++i) {
                // Transfer data on socket
                Info& p = *workers[i];
                if (FD_ISSET(p.fd, &rset)) {
                    int n;
                    while ((n = read(p.fd, buffer, sizeof(buffer))) > 0) {
                        p.worker->handleRead(buffer, n);
                    }
                    if (n == 0) {
                        p.worker->markEndOfInput();
                        p.status |= EndOfInput;
                    }
                }
                if (FD_ISSET(p.fd, &wset)) {
                    std::size_t n = p.worker->getBytesToSend(ptr);
                    if (n > 0) {
                        int did = write(p.fd, ptr, n);
                        if (did > 0) {
                            p.worker->handleWrite(did);
                        } else if (did == 0) {
                            printf("[io] [fd:%d] [id:%d] write returned 0, wtf?\n", p.fd, p.worker->getId());
                            p.status |= EndOfOutput;
                        } else {
                            MutexGuard g(output_mutex);
                            printf("[io] [fd:%d] [id:%d] write error %d (%s)\n", p.fd, errno, strerror(errno), p.worker->getId());
                            p.status |= EndOfOutput;
                        }
                    }
                }
            }
        }
    }
}

void
IOThread::processFinishedWorkers()
{
    std::size_t i = workers.size();
    while (i > 0) {
        --i;
        Info& p = *workers[i];
        if ((p.status & EndOfOutput) != 0 || p.worker->isTerminated()) {
            // Log it
            {
                MutexGuard g(output_mutex);
                printf("[io] [fd:%d] [id:%d] closing\n", p.fd, p.worker->getId());
            }

            // Close the fd
            close(p.fd);

            // This worker is terminated. Swap it to the end and drop it.
            workers.swap_elems(i, workers.size()-1);
            workers.pop_back();
        }
    }
}

void
IOThread::processNewWorkers()
{
    MutexGuard g(mutex);
    while (!new_workers.empty()) {
        std::auto_ptr<Info> p(new_workers.extract_last());
        int fd = p->fd;
        int id = p->worker->getId();
        if (fd >= FD_SETSIZE) {
            // Cannot use this!
            close(fd);

            // Log it
            MutexGuard g(output_mutex);
            printf("[io] [fd:%d] [error] file descriptor cannot be handled\n", fd);
        } else {
            // Transfer it to work list
            workers.push_back(p.release());

            // Log it
            MutexGuard g(output_mutex);
            printf("[io] [fd:%d] [id:%d] new\n", fd, id);
        }
    }
}
