/**
  *  \file planetscentral/talk/cmdrender.cc
  *  \brief Talk: RENDER commands
  */

#include "planetscentral/talk/cmdrender.h"
#include "planetscentral/talk/talkconnection.h"
#include "planetscentral/talk/render.h"
#include "planetscentral/dbclient/reconnect.h"
#include "planetscentral/talk/talk.h"
#include "planetscentral/dbclient/dbclient.h"

/* @q RENDEROPTION [FORMAT fmt:Str] [BASEURL url:Str] (Talk Command)
   Set renderer options.
   Options are used for all future rendering jobs on this connection that do not specify an override.

   The %url is used to generate links.

   The %fmt can be one of:
   - raw: do not render; just produce the raw {@type TalkText}.
   - format: do not render; just produce the type of the {@type TalkText}.
   - html: render HTML.
   - mail: render into internet email.
   - news: render into a Usenet posting.
   - text: produce just the raw text.
   - format<em>LS</em>: produce BBcode (with auto-link, auto-smiley option as given).
   If %fmt equals the format of the original {@type TalkText},
   the original text is returned as-is.

   %fmt can also contain modifiers:
   - quote:<em>format</em>: quote the text and render it in the given format.
   - noquote:<em>format</em>: remove all quotes and render the result in the given format.
   - break:<em>format</em>: render only up to the given break indicator.
   - abstract:<em>format</em>: render an abstract.
   - force:<em>format</em>: force rendering even if input and output format are the same.

   Permissions: none. */
void
doRenderOptions(TalkConnection& conn, IntArgBlock& args)
{
    while (args.getNumArgs() > 0) {
        string_t optUC = strUCase(toStringResult(args.getNext()));
        if (!conn.getRenderState().handleOption(optUC, args)) {
            conn.sendError("400 Invalid option");
            return;
        }
    }
    conn.sendSuccess("OK");
}

/* @q RENDER text:TalkText [renderOptions...] (Talk Command)
   Render text.

   The message is rendered using the current render options, see {RENDEROPTION}.
   You can temporarily override rendering options by specifying the new settings within the command.

   Permissions: none.

   @retval Str rendered text */
void
doRender(TalkConnection& conn, IntArgBlock& args)
{
    // Get text to render
    args.checkArgumentCountAtLeast(1);
    string_t text = toStringResult(args.getNext());

    // Get options
    RenderState state(conn.getRenderState());
    while (args.getNumArgs() > 0) {
        string_t optUC = strUCase(toStringResult(args.getNext()));
        if (!state.handleOption(optUC, args)) {
            conn.sendError("400 Invalid option");
            return;
        }
    }

    // Do it
    ReconnectingGuard g(database_rc);
    IntStringValue sv(state.render(text, database_connection));
    conn.sendObject(&sv);
}
